----------------------------------------------------------------------
-- Finite Field Chapter 5 package body (finite_field_ch5.adb)
-- 
-- Defines useful funtions for finite field operations
-- of Chapter 5 examples
---------------------------------------------------------------------
package body finite_fields_ch5 is

function mod_m_addition(x, y, m, k: natural) return natural is
z1, z2, c1, c2: natural;
begin
z1 := x + y; z2 := (z1 mod 2**k) + (2**k - m);
c1 := z1/2**k; c2 := z2/2**k; 
if c1 = 0 and c2 = 0 then return (z1 mod 2**k); else return (z2 mod 2**k); end if;
end mod_m_addition;

-------------------------------------------------------

function mod_m_subtraction(x, y, m, k: natural) return natural is
z1, z2, c1, sum: natural;
begin
sum := x + (2**k - y);
z1 := (sum mod 2**k); c1 := sum/2**k;
z2 := (z1 + m) mod 2**k;
if c1 = 1 then return z1; else return z2; end if;
end mod_m_subtraction;

-------------------------------------------------------

function dar_mod_multiplication(x, y, m, k: natural) return natural is 
p, q, product: natural;
z: bit_vector(0 .. k-1);
begin
q := x;
for i in 0 .. k-1 loop z(i) := q mod 2; q := q/2; end loop; 
p := 0;
for i in 0 .. k-1 loop
   p := mod_m_addition(p, p, m, k);
   if z(k-i-1) = 1 then p := mod_m_addition(p, y, m, k); end if;
end loop;
product := p;
return product;
end dar_mod_multiplication;

--------------------------------------------------------

function quotient(s: in integer; y: in natural) return integer is
begin
if s < 0 then return -1; else return 1; end if;
end quotient;

--------------------------------------------------------

function nr_reducer(x: in integer; p: in natural; n,k: in natural) return integer is
y, s, r, z: integer;
begin
y := p*(2**(n-k));
s := x;
for i in 0 .. n-k loop
   if quotient(s,y) = 1 then r := s - y;
   elsif quotient(s,y) = 0 then r := s;
   else r := s + y;
   end if;
   s := 2*r;
end loop;
z := r / (2**(n-k));
if z < 0 then z := (z + p); end if;
return z;
end nr_reducer;


end finite_fields_ch5;

