----------------------------------------------------------------------------
-- MSE first Mod f multiplier (MSE_first_Mod_f_Multiplier.vhd)
--
-- The hardware is genenerate for a specific P = 239.
-- And for an especified F
-- use Most Significant Element (MSE) First algorithm
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package mod_f_multiplier_parameters_package is
  constant K: natural := 8;
  constant P: std_logic_vector(k-1 downto 0) := conv_std_logic_vector(239, k);
  constant M: natural := 17;
  type long_polynomial is array(M downto 0) of std_logic_vector(K-1 downto 0);
  type polynomial is array(M-1 downto 0) of std_logic_vector(K-1 downto 0);
  constant LOGM: natural := 5;--logm is the number of bits of m-1
  constant f: long_polynomial :=
  ("00000001", "00000000", "00000000", "00000000", "00000000", "00000000",
   "00000000", "00000000", "00000000", "00000000", "00000000", "00000000",
   "00000000", "00000000", "00000000", "00000000", "00000000", "11101101");
  --("00000001", "00000000", "00000000", "00000000", "00010001", "00000000",
  --"00011101", "00000000", "11000010", "00000000", "00000000", "00000000",
  --"00000000", "00000000", "00000000", "00000000", "00000000", "11101101" );
  constant zero_coef: std_logic_vector(K-1 downto 0) := (others => '0');
  constant zero_poly: polynomial := (others => zero_coef);
end mod_f_multiplier_parameters_package;

------------------------------------------------------------
-- mod 239 reducer
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_reducer is
port (
  x: in std_logic_vector(15 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_reducer;

architecture circuit of mod_239_reducer is
  signal x1_by_17, x0, sum, xx1_by_17, xx0, xxx: std_logic_vector(8 downto 0);
  signal long_xxx, minus_239, dif: std_logic_vector(9 downto 0);
  signal x2_by_33, long_sum, xx: std_logic_vector(9 downto 0);
begin
  x1_by_17 <= '0'&x(11 downto 8)&x(11 downto 8);
  x0 <= '0'&x(7 downto 0);
  sum <= x1_by_17 + x0;
  x2_by_33 <= '0'&x(15 downto 12)&'0'&x(15 downto 12);
  long_sum <= '0'&sum;
  xx <= x2_by_33 + long_sum;
  xx1_by_17 <= "000"&xx(9 downto 8) &"00"&xx(9 downto 8);
  xx0 <= '0'&xx(7 downto 0);
  xxx <= xx1_by_17 + xx0;
  minus_239 <= conv_std_logic_vector(273, 10);
  long_xxx <= '0'&xxx;
  dif <= long_xxx + minus_239;
  with dif(9) select z <= dif(7 downto 0) when '1', xxx(7 downto 0) when others;
end circuit;

------------------------------------------------------------
-- mod 239 Multiplier
------------------------------------------------------------
library ieee; use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
entity mod_239_multiplier is
port (
  x, y: in std_logic_vector(7 downto 0);
  z: out std_logic_vector(7 downto 0)
  );
end mod_239_multiplier;

architecture circuit of mod_239_multiplier is
  component mod_239_reducer is
  port (
    x: in std_logic_vector(15 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  signal product: std_logic_vector(15 downto 0);
begin
  product <= x*y;
  main_component: mod_239_reducer port map(product, z);
end circuit;

------------------------------------------------------------
-- adder - subtractor 
------------------------------------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.mod_f_multiplier_parameters_package.all;
entity adder_subtractor is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  addb_sub: in std_logic;
  z: out std_logic_vector(k-1 downto 0)
  );
end adder_subtractor;

architecture rtl of adder_subtractor is
  signal long_x, xor_y, sum1, long_z1, xor_p, sum2: std_logic_vector(k downto 0);
  signal c1, c2, sel: std_logic;
  signal z1, z2: std_logic_vector(k-1 downto 0);
begin
  long_x <= '0'&x;
  xor_gates1: for i in 0 to k-1 generate xor_y(i) <= y(i) xor addb_sub; end generate;
  xor_y(k) <= '0';
  sum1 <= addb_sub + long_x + xor_y;
  c1 <= sum1(k);
  z1 <= sum1(k-1 downto 0);
  long_z1 <= '0'&z1;
  xor_gates2: for i in 0 to k-1 generate xor_p(i) <= p(i) xor not(addb_sub); end generate;
  xor_p(k) <= '0';
  sum2 <= not(addb_sub) + long_z1 + xor_p;
  c2 <= sum2(k);
  z2 <= sum2(k-1 downto 0);
  sel <= (not(addb_sub) and (c1 or c2)) or (addb_sub and not(c1));
  with sel select z <= z1 when '0', z2 when others;
end rtl;

------------------------------------------------------------
-- mod f multiplier 
------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.mod_f_multiplier_parameters_package.all;
entity MSE_first_mod_f_multiplier is
port(
  a, b: in polynomial;
  clk, reset, start: in std_logic;  
  z: out polynomial;
  done: out std_logic
  );
end MSE_first_mod_f_multiplier;

architecture circuit of MSE_first_mod_f_multiplier is
  component mod_239_multiplier is
  port(
    x, y: in std_logic_vector(7 downto 0);
    z: out std_logic_vector(7 downto 0) );
  end component;
  component adder_subtractor is
  port (
    x, y: in std_logic_vector(k-1 downto 0);
    addb_sub: in std_logic;
    z: out std_logic_vector(k-1 downto 0) );
  end component;
  signal c, add_in1, add_in2, mult_in1, mult_in2, next_c, int_b: polynomial;
  signal serial_out: std_logic_vector(k-1 downto 0); 
  signal sel, ce_c, load, update, count_equal_zero: std_logic;
  signal count: std_logic_vector(logm-1 downto 0);
  type states is range 0 to 4;
  signal current_state: states;

begin

  main_component: for i in 1 to m-1 generate
    with sel select add_in1(i) <= c(i) when '0', c(i-1) when others;
    with sel select mult_in1(i) <= c(m-1) when '1', serial_out when others;
    with sel select mult_in2(i) <= f(i) when '1', a(i) when others;
    comp1: mod_239_multiplier port map(mult_in1(i), mult_in2(i), add_in2(i));
    comp2: adder_subtractor port map(add_in1(i), add_in2(i), sel, next_c(i));
  end generate;
  with sel select add_in1(0) <= c(0) when '0', zero_coef when others;
  with sel select mult_in1(0) <= c(m-1) when '1', serial_out when others;
  with sel select mult_in2(0) <= f(0) when '1', a(0) when others;
  comp1: mod_239_multiplier port map(mult_in1(0), mult_in2(0), add_in2(0));
  comp2: adder_subtractor port map(add_in1(0), add_in2(0), sel, next_c(0));

  c_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then c <= zero_poly;
    elsif ce_c = '1' then c <= next_c;
    end if;
  end if;
  end process c_register;
  z <= c;

  shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then int_b <= b;
    elsif update = '1' then
      for i in m-1 downto 1 loop int_b(i) <= int_b(i-1); end loop;
      int_b(0) <= zero_coef;
    end if;
  end if;
  end process shift_register;

  serial_out <= int_b(m-1);

  a_counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then count <= 
      conv_std_logic_vector(M-1, LOGM);
    elsif update = '1' then 
      count <= count - 1;
    end if;
  end if;
  end process a_counter;

  count_equal_zero <= '1' when count = 0 else '0';

  control_unit: process(clk, reset, count_equal_zero, current_state)
  begin
  case current_state is
    when 0 to 1 => sel <= '0'; ce_c <= '0'; load <= '0'; update <= '0'; done <= '1';
    when 2 => sel <= '0'; ce_c <= '0'; load <= '1'; update <= '0'; done <= '0';
    when 3 => sel <= '1'; ce_c <= '1'; load <= '0'; update <= '0'; done <= '0';
    when 4 => sel <= '0'; ce_c <= '1'; load <= '0'; update <= '1'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => current_state <= 4;
      when 4 => if count_equal_zero = '1' then current_state <= 0; else current_state <= 3; end if;
    end case;
  end if;

  end process control_unit;

end circuit;
