----------------------------------------------------------------------------
-- Adder of Polynomials (adder_polynom.vhd)
--
-- Adder mod p in Zp[x] / f(x)
-- The hardware is genenerate for a specific P.
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package adder_subt_parameters is
  constant K: natural := 5; --number of bits of P
  constant M: natural := 8;
  constant P: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(17, K);
--  constant K: natural := 8;
--  constant M: natural := 17;
--  constant P: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(239, K);
  type polynomial is array(M-1 downto 0) of std_logic_vector(K-1 downto 0);
end adder_subt_parameters;

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.adder_subt_parameters.all;

entity adder_polynom is
port(
  a, b: in polynomial;
  z: out polynomial
);
end adder_polynom;

architecture circuit of adder_polynom is
begin

  gen: for i in 0 to M-1 generate
    addition: process(a,b)
      variable z1, z2: std_logic_vector(K downto 0);
    begin
      z1 := a(i) +  ('0' & b(i));
      z2 := z1 - P;
      if z1(K) = '0' then z(i) <= z1(K-1 downto 0); 
      else z(i) <= z1(K-1 downto 0); end if;
    end process;
  end generate;
end circuit;


