----------------------------------------------------------------------------
-- Adder/Subtractor (adder_subtractor_polynom.vhd)
--
-- Adds or subtract mod p in Zp[x] / f(x)
-- The hardware is genenerate for a specific p.
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package adder_subt_parameters is
  constant K: natural := 5; --number of bits of P
  constant M: natural := 8;
  constant P: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(17, K);
--  constant K: natural := 8;
--  constant M: natural := 17;
--  constant P: std_logic_vector(K-1 downto 0) := conv_std_logic_vector(239, K);
  type long_polynomial is array(M downto 0) of std_logic_vector(K-1 downto 0);
  type polynomial is array(M-1 downto 0) of std_logic_vector(K-1 downto 0);
  constant zero_coef: std_logic_vector(K-1 downto 0) := (others => '0');
  constant zero_poly: polynomial := (others => zero_coef);
end adder_subt_parameters;

library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.adder_subt_parameters.all;

entity adder_subtractor is
port (
  x, y: in std_logic_vector(K-1 downto 0);
  add_sub: in std_logic;
  z: out std_logic_vector(K-1 downto 0)
);
end adder_subtractor;

architecture rtl of adder_subtractor is
  signal long_x, xor_y, sum1, long_z1, xor_p, sum2: std_logic_vector(K downto 0);
  signal c1, c2, sel: std_logic;
  signal z1, z2: std_logic_vector(K-1 downto 0);
begin
  long_x <= '0' & x;
  xor_gates1: for i in 0 to K-1 generate 
    xor_y(i) <= y(i) xor add_sub; 
  end generate;
  xor_y(K) <= '0';
  sum1 <= add_sub + long_x + xor_y;
  c1 <= sum1(K);
  z1 <= sum1(K-1 downto 0);
  long_z1 <= '0'&z1;
  xor_gates2: for i in 0 to K-1 generate 
    xor_p(i) <= P(i) xor not(add_sub); 
  end generate;
  xor_p(K) <= '0';
  sum2 <= not(add_sub) + long_z1 + xor_p;
  c2 <= sum2(K);
  z2 <= sum2(K-1 downto 0);
  sel <= (not(add_sub) and (c1 or c2)) or (add_sub and not(c1));
  with sel select z <= z1 when '0', z2 when others;
end rtl;


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.adder_subt_parameters.all;

entity add_sub_polynom is
port(
  a, b: in polynomial;
  add_sub: in std_logic;  
  z: out polynomial
);
end add_sub_polynom;

architecture circuit of add_sub_polynom is
  component adder_subtractor is
  port (
    x, y: in std_logic_vector(K-1 downto 0);
    add_sub: in std_logic;
    z: out std_logic_vector(K-1 downto 0)
  );
  end component;
begin

  main_component: for i in 0 to M-1 generate
    comp1: adder_subtractor port map(x => a(i), y => b(i),
                                      add_sub => add_sub, z => z(i));
  end generate;
end circuit;


