--------------------------------------------------------------------------------
--  Test for Polynomial multiplier (ch5)
--
--  Generates random vectors and comparte results 
--
--
--------------------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_textio.ALL;
use ieee.math_real.all; -- for UNIFORM, TRUNC
use work.mod_f_multiplier_parameters_package.all;
USE std.textio.ALL;

ENTITY test_mult_polynom IS
END test_mult_polynom;

ARCHITECTURE behavior OF test_mult_polynom IS 

  -- Component Declaration for the Unit Under Test (UUT1)
  component LSE_first_mod_f_multiplier is
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component;

  component MSE_first_mod_f_multiplier is
  port(
    a, b: in polynomial;
    clk, reset, start: in std_logic;  
    z: out polynomial;
    done: out std_logic );
  end component;
  
    -- Internal signals
  SIGNAL clk, reset, start, done_mse, done_lse: std_logic;
  constant DELAY : time := 100 ns;
  constant PERIOD : time := 200 ns;
  constant DUTY_CYCLE : real := 0.5;
  constant OFFSET : time := 0 ns;
  constant NUMBER_TESTS: natural := 10;
  
  signal x, y, z_lse, z_mse: polynomial := (others => zero_coef);

BEGIN

  -- Instantiate the Unit Under Test (UUT)
  uut1: MSE_first_mod_f_multiplier port map ( a => x, b => y, 
            clk => clk, reset => reset, start => start, z => z_mse, done => done_mse);  
  uut2: LSE_first_mod_f_multiplier port map ( a => x, b => y, 
            clk => clk, reset => reset, start => start, z => z_lse, done => done_lse);  

  PROCESS -- clock process for clk
  BEGIN
  WAIT for OFFSET;
  CLOCK_LOOP : LOOP
     clk <= '0';
     WAIT FOR (PERIOD *(1.0 - DUTY_CYCLE));
     clk <= '1';
     WAIT FOR (PERIOD * DUTY_CYCLE);
  END LOOP CLOCK_LOOP;
  END PROCESS;

  tb_proc : PROCESS --generate values
    PROCEDURE gen_polynom(X : out polynomial; w: natural; s1, s2: inout Natural) IS
      VARIABLE i_x, i_p: natural;
      VARIABLE rand: real;
    BEGIN
        i_p := conv_integer(P);
        for i in 0 to M-1 loop
          UNIFORM(s1, s2, rand);
          i_x := INTEGER(TRUNC(rand * real(i_p)));
          x(i) := CONV_STD_LOGIC_VECTOR (i_x, K);
        end loop;
    END PROCEDURE;

    VARIABLE seed1, seed2: positive; 
    VARIABLE i_x, i_y, i_p, i_z, i_yz_modp: integer;
    VARIABLE cycles, max_cycles, min_cycles, total_cycles: integer := 0;
    VARIABLE avg_cycles: real;
    VARIABLE initial_time, final_time: time;
    VARIABLE xx: polynomial ;

  BEGIN
    start <= '0'; reset <= '1';
    WAIT FOR PERIOD;
    reset <= '0';
    WAIT FOR PERIOD;
    
    for I in 1 to NUMBER_TESTS loop
      gen_polynom(xx, M, seed1, seed2);
      x <= xx;
      gen_polynom(xx, M, seed1, seed2);
      y <= xx;
      
      start <= '1'; 
      WAIT FOR PERIOD;
      start <= '0';
      wait until done_mse = '1';

      WAIT FOR 2*PERIOD;
      
      IF ( z_lse /= z_mse ) THEN 
        ASSERT (FALSE) REPORT "ERROR!!! Z for LSE different than for MSE method" SEVERITY ERROR;
      END IF;  
 
    end loop;
    WAIT FOR DELAY;
    ASSERT (FALSE) REPORT "Simulation successful!.  Not error detected"
    SEVERITY FAILURE;
  END PROCESS;

END;
